import openai
from langchain.embeddings import OpenAIEmbeddings
import pandas as pd
import os
import json
import concurrent.futures
from threading import Lock
from termcolor import colored

import tenacity

import logging
logger = logging.getLogger(__name__)

# Setting up API key and embedding model
openai.api_key = os.getenv("OPENAI_API_KEY")

# implement exponential backoff, try 10 times, wait 1 second after the first failure, then 2 seconds, then 4 seconds, then 8 seconds, then 16 seconds, then 3
@tenacity.retry(wait=tenacity.wait_exponential(multiplier=2, min=1, max=60), stop=tenacity.stop_after_attempt(10), after=tenacity.after_log(logger, logging.DEBUG))
def get_embedding(text: str) -> list:
    """Embed the text using the provided embeddings model."""
    embeddings_model = OpenAIEmbeddings()
    res = embeddings_model.embed_documents([text])[0]
    return res

# implement exponential backoff, try 10 times, wait 1 second after the first failure, then 2 seconds, then 4 seconds, then 8 seconds, then 16 seconds, then 3
@tenacity.retry(wait=tenacity.wait_exponential(multiplier=2, min=1, max=60), stop=tenacity.stop_after_attempt(10), after=tenacity.after_log(logger, logging.DEBUG))
def generate_system_message(prompt):

    response = openai.ChatCompletion.create(
        model="gpt-4",
        messages=[
          {
            "role": "system",
            "content": "You will be given a high-level description of the model we are training, and from that, you will generate a simple system prompt for that model to use. Remember, you are not generating the system message for data generation -- you are generating the system message to use for inference. A good format to follow is `Given $INPUT_DATA, you will $WHAT_THE_MODEL_SHOULD_DO.`.\n\nMake it as concise as possible. Include nothing but the system prompt in your response.\n\nFor example, never write: `\"$SYSTEM_PROMPT_HERE\"`.\n\nIt should be like: `$SYSTEM_PROMPT_HERE`."
          },
          {
              "role": "user",
              "content": prompt.strip(),
          }
        ],
        temperature=temperature,
        max_tokens=500,
    )

    return response.choices[0].message['content']
# Configuration constants
# we construct the dataset prompt by combining a description of the prompt with the desired response prompt so that we can inser the desired response prompt into function description
# full dataset prompt: "A model that takes in a prompt in English for a python function, and responds with the simplest, smallest, fully functional, most efficient python function that satisfies the prompt."
desired_response_prompt = "the simplest, smallest, fully functional, most efficient python function that satisfies the prompt."
dataset_prompt =  f"A model that takes in a prompt in English for a python function, and responds with {desired_response_prompt}"
model = "gpt-4"
temperature = .4
topic_counts = {}
# total number of examples to generate. you can always keep running this script to generate more examples and add them to the datasets
# LARGER THE NUMBER OF EXAMPLES, THE MORE IT WILL COST!!!!!
number_of_examples = 50
total_examples_per_topic = 10

# generate a new system message dataset if we dont have one in a file
# !!!!! SYSTEM MESSAGE WILL BE USED FROM THE FILE IF IT EXISTS, THIS IS ONLY USEFUL IF YOU WANT TO CONTINUE GENERATING EXAMPLES FOR THE SAME SYSTEM MESSAGE !!!!!
print(colored("IF 'SYSTEM_MESSAGE_DATASET.TXT' EXISTS, SYSTEM MESSAGE WILL BE LOADED FROM THERE. IF YOU WANT TO GENERATE A NEW SYSTEM MESSAGE, DELETE THAT FILE. BEFORE PROCEEDING", "red"))
USER_INPUT = input("Press enter to continue or type 'exit' to exit\n")
if USER_INPUT == "exit":
    exit()
if not os.path.exists("system_message_dataset.txt"):
    while True:
        system_message_dataset = generate_system_message(dataset_prompt)
        user_confirmation = input(f"Do you accept the following system message?, 'y' to accept, 'n' to generate a new one! (y/n)\n{system_message_dataset}\n")
        if user_confirmation == "y":
            # write the system message to a file
            with open("system_message_dataset.txt", "w", encoding="utf-8") as f:
                f.write(system_message_dataset)
        break
else:
    # load the system message from the file
    with open("system_message_dataset.txt", "r", encoding="utf-8") as f:
        system_message_dataset = f.read()


# Load the dataset if it exists
if os.path.exists("dataset_with_embeddings.csv"):
    runtime_df = pd.read_csv("dataset_with_embeddings.csv")
else:
    # initialize the runtime dataframe if it doesn't exist
    runtime_df = pd.DataFrame(columns=["prompt", "response", "embedding"])


# this function updates the topic counts and saves them to a file
def update_topic_counts(current_covered_topic):
    global topic_counts
    # load from file if it exists
    if os.path.exists("topic_counts.json"):
        with open("topic_counts.json", "r", encoding="utf-8") as f:
            topic_counts = json.load(f)
    if current_covered_topic is None:
        return
        
    if current_covered_topic in topic_counts:
        topic_counts[current_covered_topic] += 1
    else:
        topic_counts[current_covered_topic] = 1
    # save the topic counts to a file
    with open("topic_counts.json", "w", encoding="utf-8") as f:
        json.dump(topic_counts, f, ensure_ascii=False, indent=4, separators=(',', ': '))

# implement exponential backoff, try 10 times, wait 1 second after the first failure, then 2 seconds, then 4 seconds, then 8 seconds, then 16 seconds, then 3
@tenacity.retry(wait=tenacity.wait_exponential(multiplier=2, min=1, max=60), stop=tenacity.stop_after_attempt(10), after=tenacity.after_log(logger, logging.DEBUG))
def generate_dataset(dataset_prompt, model):
    update_topic_counts(current_covered_topic=None)
    """Generate an example dataset using the provided model and prompt."""
    # Creating the API call's message structure
    base_message = {
        "role": "system",
        "content": f"You are generating a diverse dataset which will be used to train a machine learning model.\n\nYou will be given a high-level description of the model we want to train, and from that, you will generate data samples, each with a prompt/response pair.For each turn, make the example slightly more complex than the last, while ensuring diversity.\n\nWe need dataset to cover a wide variety of topics and which is in the model's domain\n\nEnsure that you diversify the dataset bychoosing different topics than most covered so far\n\nMake sure your samples are unique and diverse, yet high-quality and complex enough to train a well-performing model.generate total of max {total_examples_per_topic} examples for a topic category\n\nHere is the type of model we want to train:`{dataset_prompt}`"
    }
    past_prompts_instruction = {"role": "user", "content": "Make sure to diversify each new example from the examples that are already in the dataset as follows:"}
    topics_content = "\n".join([f"{topic}: {count} times" for topic, count in topic_counts.items()])
    # print("Topics covered so far:", topics_content)
    assistant_messages = [{"role": "assistant", "content":"diversify in a different direction than most covered topics. Add to less covered methods to create a bias free distribution. Here are the topics covered so far:\n\n" + topics_content}]

    
    messages = [base_message] + [past_prompts_instruction] + assistant_messages

    response = openai.ChatCompletion.create(
        model=model,
        messages=messages,
        stream=True,
        temperature=temperature,
        functions = [
            {
                "name": "dataset_item",
                "description": "A dataset item is a prompt/response pair.",
                "parameters": {
                    "type": "object",
                    "properties": {

                        "diversification_direction": {
                            'type': 'string',
                            'description': "a detailed description of a more complex and diverse topic to diversify to which is different from the topics covered so far in the dataset.",
                        },
                        "prompt": {
                            "type": "string",
                            "description": "The diversified prompt to generate a response for."
                        },
                        "response": {
                            "type": "string",
                            "description": desired_response_prompt,
                        },
                        "current_covered_topic": {
                            "type": "string",
                            "description": "topic of the example currently being generated."
                        }
                    },
                    "required": ["prompt", "response", "diversification_direction", "current_covered_topic"]
                }
            }
        ],
        function_call = {"name": "dataset_item"} # can be "auto" or "none"
    )
    

    responses = ''
    for chunk in response:
        if chunk["choices"][0]["delta"].get("function_call"):
            chunk = chunk["choices"][0]["delta"]
            processed_text = chunk["function_call"]["arguments"]
            # print(processed_text, end='', flush=True)
            responses += processed_text

    return responses

def generate_example(dataset_prompt, model_name):
    """Extract the prompt and response from the generated dataset."""
    response = generate_dataset(dataset_prompt, model_name)

    try:
        new_prompt = json.loads(response)["prompt"]
        new_response = json.loads(response)["response"]
        diversity_direction = json.loads(response)["diversification_direction"]
        current_covered_topic = json.loads(response)["current_covered_topic"]
        update_topic_counts(current_covered_topic)
        print(colored("Current covered topic:", "green"), current_covered_topic)

        print("Diversification direction:", diversity_direction)
        
        combined_response = "# " + new_prompt + "\n\n" + new_response + "\n\n"
        # create dataset_files folder if it doesn't exist
        if not os.path.exists("dataset_files"):
            os.mkdir("dataset_files")
        # save the combined response to a file
        with open(f"dataset_files/{new_prompt}.py", "a", encoding="utf-8") as f:
            f.write(combined_response)
        
        return (new_prompt, new_response, combined_response)
    except:
        print("Error: response is not valid JSON")
        return None, None, None

def generate_dataframe(dataset_prompt, model_name):
    """Generate a dataframe from the extracted prompt and response."""
    new_prompt, new_response, combined_response = generate_example(dataset_prompt, model_name)

    if new_prompt and new_response:
        df = pd.DataFrame({
            "prompt": [new_prompt],
            "response": [new_response],
            "embedding": [get_embedding(combined_response)]
        })
        # drop identical rows based on prompt 
        return df
    return None

file_lock = Lock()

def save_to_files(df):
    """Save the dataframe to various file formats."""
    global runtime_df

    with file_lock:
        # Only keep rows from the new df that are not already in the runtime_df
        df = df[~df["prompt"].isin(runtime_df["prompt"])]
        
        if not df.empty:
            # Update the runtime dataframe with new data
            runtime_df = pd.concat([runtime_df, df], ignore_index=True)
            # At this point, the runtime_df should be free from duplicates
            
            # Save to CSV
            if os.path.exists("dataset_with_embeddings.csv"):
                df.to_csv("dataset_with_embeddings.csv", mode="a", header=False, index=False)
            else:
                df.to_csv("dataset_with_embeddings.csv", index=False)

            # Save to JSON
            if os.path.exists("dataset_with_embedding.json"):
                temp_df = pd.read_json("dataset_with_embedding.json", orient="records")
                combined_df = pd.concat([temp_df, df], ignore_index=True)
                combined_df.to_json("dataset_with_embedding.json", orient="records")
            else:
                df.to_json("dataset_with_embedding.json", orient="records")

            # Save to JSONL
            with open("finetuning_dataset.jsonl", "a", encoding="utf-8") as f:
                for index, row in df.iterrows():
                    json.dump({"messages": [{"role": "system", "content": system_message_dataset}, {"role": "user", "content": row["prompt"]}, {"role": "assistant", "content": row["response"]}]}, f, ensure_ascii=False)
                    f.write("\n")


def process_and_save(dataset_prompt, model_name):
    """Process and save the dataframe."""
    df = generate_dataframe(dataset_prompt, model_name)
    if df is not None:
        save_to_files(df)

def run_with_threadpool(number_of_examples):
    """Use threading to generate and save examples."""
    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        # Use the map function to process the tasks in parallel
        list(executor.map(process_and_save, [dataset_prompt]*number_of_examples, [model]*number_of_examples))

run_with_threadpool(number_of_examples)

